import requests
from typing import List, Optional
from sempy_labs._a_lib_info import lib_name, lib_version
from sempy._utils._log import log


@log
def _format_dax(
    expressions: str | List[str],
    skip_space_after_function_name: bool = False,
    metadata: Optional[List[dict]] = None,
) -> List[str]:

    if isinstance(expressions, str):
        expressions = [expressions]
        metadata = [metadata] if metadata else [{}]

    # Add variable assignment to each expression
    expressions = [f"x :={item}" for item in expressions]

    url = "https://daxformatter.azurewebsites.net/api/daxformatter/daxtextformatmulti"

    payload = {
        "Dax": expressions,
        "MaxLineLength": 0,
        "SkipSpaceAfterFunctionName": skip_space_after_function_name,
        "ListSeparator": ",",
        "DecimalSeparator": ".",
    }

    headers = {
        "Accept": "application/json, text/javascript, */*; q=0.01",
        "Accept-Encoding": "gzip,deflate",
        "Accept-Language": "en-US,en;q=0.8",
        "Content-Type": "application/json; charset=UTF-8",
        "Host": "daxformatter.azurewebsites.net",
        "Expect": "100-continue",
        "Connection": "Keep-Alive",
        "CallerApp": lib_name,
        "CallerVersion": lib_version,
    }

    response = requests.post(url, json=payload, headers=headers)
    result = []
    for idx, dax in enumerate(response.json()):
        formatted_dax = dax.get("formatted")
        errors = dax.get("errors")
        if errors:
            meta = metadata[idx] if metadata and idx < len(metadata) else {}
            obj_name = meta.get("name", "Unknown")
            table_name = meta.get("table", "Unknown")
            obj_type = meta.get("type", "Unknown")
            if obj_type == "calculated_tables":
                raise ValueError(
                    f"DAX formatting failed for the '{obj_name}' calculated table: {errors}"
                )
            elif obj_type == "calculated_columns":
                raise ValueError(
                    f"DAX formatting failed for the '{table_name}'[{obj_name}] calculated column: {errors}"
                )
            elif obj_type == "calculation_items":
                raise ValueError(
                    f"DAX formatting failed for the '{table_name}'[{obj_name}] calculation item: {errors}"
                )
            elif obj_type == "measures":
                raise ValueError(
                    f"DAX formatting failed for the '{obj_name}' measure: {errors}"
                )
            elif obj_type == "rls":
                raise ValueError(
                    f"DAX formatting failed for the row level security expression on the '{table_name}' table within the '{obj_name}' role: {errors}"
                )
            else:
                NotImplementedError()
        else:
            if formatted_dax.startswith("x :="):
                formatted_dax = formatted_dax[4:]
            formatted_dax = formatted_dax.strip()
            result.append(formatted_dax)
    return result



from sempy_labs._helper_functions import (
    _base_api,
    resolve_workspace_id,
    resolve_item_id,
)
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
from sempy._utils._log import log


@log
def execute_query(
    dataflow: str | UUID,
    query_name: str,
    custom_mashup_document: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Executes a query against a dataflow and returns the result.

    This is a wrapper function for the following API: `Query Execution - Execute Query <https://learn.microsoft.com/rest/api/fabric/dataflow/query-execution/execute-query>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataflow : str | uuid.UUID
        Name or ID of the dataflow.
    query_name : str
        The name of the query to execute from the dataflow (or from the custom mashup document if provided).
    custom_mashup_document : str, default=None
        Optional custom mashup document to override the dataflow's default mashup.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all parameters defined in the specified Dataflow.
    """

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=dataflow, type="Dataflow", workspace=workspace_id)

    payload = {
        "queryName": query_name,
    }

    if custom_mashup_document:
        payload["customMashupDocument"] = custom_mashup_document

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/dataflows/{item_id}/executeQuery",
        method="post",
        payload=payload,
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} The '{query_name}' query within the '{dataflow}' dataflow has been executed."
    )



from typing import Optional, List
from uuid import UUID
from sempy._utils._log import log
import pandas as pd
from sempy_labs._helper_functions import (
    resolve_lakehouse_name_and_id,
    resolve_workspace_id,
    resolve_lakehouse_id,
    _create_dataframe,
    _base_api,
    resolve_workspace_name_and_id,
)
import sempy_labs._icons as icons


@log
def is_schema_enabled(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> bool:
    """
    Indicates whether a lakehouse has schemas enabled.

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    bool
        Indicates whether the lakehouse has schemas enabled.
    """
    workspace_id = resolve_workspace_id(workspace)
    (item_name, item_id) = resolve_lakehouse_name_and_id(lakehouse, workspace)
    response = _base_api(f"/v1/workspaces/{workspace_id}/lakehouses/{item_id}")
    default_schema = response.json().get("properties", {}).get("defaultSchema", None)
    if default_schema:
        return True
    else:
        return False


@log
def list_schemas(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Lists the schemas within a Fabric lakehouse.

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows the schemas within a lakehouse.
    """

    columns = {
        "Schema Name": "str",
    }
    df = _create_dataframe(columns=columns)
    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_lakehouse_id(lakehouse, workspace)
    response = _base_api(
        request=f"{workspace_id}/{item_id}/api/2.1/unity-catalog/schemas?catalog_name={item_id}",
        client="onelake",
    )

    rows = []
    for s in response.json().get("schemas", []):
        rows.append(
            {
                "Schema Name": s.get("name", None),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


def list_tables(
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    schema: Optional[str | List[str]] = None,
) -> pd.DataFrame:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_lakehouse_name_and_id(lakehouse, workspace)

    response = _base_api(f"/v1/workspaces/{workspace_id}/lakehouses/{item_id}")
    default_schema = response.json().get("properties", {}).get("defaultSchema", None)
    schema_enabled = True if default_schema else False

    columns = {
        "Workspace Name": "str",
        "Lakehouse Name": "str",
        "Table Name": "str",
        "Schema Name": "str",
        "Format": "str",
        "Type": "str",
        "Location": "str",
    }
    df = _create_dataframe(columns=columns)

    prefix = "onelake.dfs.fabric.microsoft.com/"
    rows = []
    if schema_enabled:
        schemas = list_schemas(lakehouse=lakehouse, workspace=workspace)
        if schema:
            if isinstance(schema, str):
                schema = [schema]
            schemas = schemas[schemas["Schema Name"].isin(schema)]

        # Loop through schemas
        for _, r in schemas.iterrows():
            schema_name = r["Schema Name"]
            response = _base_api(
                request=f"{workspace_id}/{item_id}/api/2.1/unity-catalog/tables?catalog_name={item_id}&schema_name={schema_name}",
                client="onelake",
            )
            # Loop through tables
            for t in response.json().get("tables", []):
                location = t.get("storage_location", {})
                # Strip protocol + host
                path = location.replace(f"https://{prefix}", "")

                # Split components
                w_id, lh_id, *rest = path.split("/")
                rest = [p for p in rest]

                # Rebuild pattern
                rebuilt_location = (
                    f"abfss://{w_id}" f"@{prefix}" f"{lh_id}/" + "/".join(rest)
                )
                rows.append(
                    {
                        "Workspace Name": workspace_name,
                        "Lakehouse Name": item_name,
                        "Table Name": t.get("name", {}),
                        "Schema Name": schema_name,
                        "Format": t.get("data_source_format", {}).lower(),
                        "Type": "Managed",
                        "Location": rebuilt_location,
                    }
                )
    else:
        if schema:
            print(
                f"{icons.info} The schema parameter has been ignored as the '{item_name}' lakehouse within the '{workspace_name}' workspace has schemas disabled."
            )
        responses = _base_api(
            request=f"v1/workspaces/{workspace_id}/lakehouses/{item_id}/tables",
            uses_pagination=True,
            client="fabric_sp",
        )
        for r in responses:
            for i in r.get("data", []):
                rows.append(
                    {
                        "Workspace Name": workspace_name,
                        "Lakehouse Name": item_name,
                        "Schema Name": None,
                        "Table Name": i.get("name"),
                        "Format": i.get("format"),
                        "Type": i.get("type"),
                        "Location": i.get("location"),
                    }
                )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


def schema_exists(
    schema: str,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
) -> bool:
    """
    Indicates whether the specified schema exists within a Fabric lakehouse.

    Parameters
    ----------
    schema : str
        The name of the schema.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    bool
        Indicates whether the specified schema exists within the lakehouse.
    """

    df = list_schemas(lakehouse=lakehouse, workspace=workspace)
    return schema in df["Schema Name"].values

    # (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    # (item_name, item_id) = resolve_lakehouse_name_and_id(lakehouse, workspace)
    # response = _base_api(
    #    request=f"{workspace_id}/{item_id}/api/2.1/unity-catalog/schemas/{schema}",
    #    client="onelake",
    #    method="head",
    # )

    # response.json()



import re
from uuid import UUID
from typing import Optional, List
from sempy_labs._helper_functions import (
    _base_api,
    resolve_workspace_name,
)
from sempy._utils._log import log


@log
def _extract_expression_list(expression):
    """
    Finds the pattern for DL/SQL & DL/OL expressions in the semantic model.
    """

    pattern_sql = r'Sql\.Database\s*\(\s*"([^"]+)"\s*,\s*"([^"]+)"\s*\)'
    pattern_no_sql = (
        r'AzureStorage\.DataLake\(".*?/([0-9a-fA-F\-]{36})/([0-9a-fA-F\-]{36})"'
    )

    match_sql = re.search(pattern_sql, expression)
    match_no_sql = re.search(pattern_no_sql, expression)

    result = []
    if match_sql:
        value_1, value_2 = match_sql.groups()
        result = [value_1, value_2, True]
    elif match_no_sql:
        value_1, value_2 = match_no_sql.groups()
        result = [value_1, value_2, False]

    return result


@log
def _get_direct_lake_expressions(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Extracts a dictionary of all Direct Lake expressions from a semantic model.
    """

    from sempy_labs.tom import connect_semantic_model

    result = {}

    with connect_semantic_model(dataset=dataset, workspace=workspace) as tom:
        for e in tom.model.Expressions:
            expr_name = e.Name
            expr = e.Expression

            list_values = _extract_expression_list(expr)
            if list_values:
                result[expr_name] = list_values

    return result


@log
def get_direct_lake_sources(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> List[dict]:
    """
    Retrieves a list of the Direct Lake sources used in a semantic model, including their type, workspace, and whether they use a SQL endpoint.

    Parameters:
    dataset : str | uuid.UUID
        The name or UUID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns:
    typing.List[dict]
        A list of dictionaries, each containing details about a Direct Lake source used in the semantic model.
        Example:
        [
            {
                "itemId": "123e4567-e89b-12d3-a456-426614174000",
                "itemName": "My Lakehouse",
                "itemType": "Lakehouse",
                "workspaceId": "123e4567-e89b-12d3-a456-426614174001",
                "workspaceName": "Workspace A",
                "usesSqlEndpoint": False,
            },
            {
                "itemId": "123e4567-e89b-12d3-a456-426614174002",
                "itemName": "My Data Source",
                "itemType": "Warehouse",
                "workspaceId": "123e4567-e89b-12d3-a456-426614174001",
                "workspaceName": "Workspace A",
                "usesSqlEndpoint": True,
            }
        ]

    """

    sql = "SqlAnalyticsEndpoint"
    sources = []
    expr = _get_direct_lake_expressions(dataset=dataset, workspace=workspace)
    for name, items in expr.items():
        artifact_id = items[1]
        uses_sql_endpoint = items[2]
        result = _base_api(
            request=f"metadata/artifacts/{artifact_id}", client="internal"
        ).json()
        item_type = result.get("artifactType")
        item_name = result.get("displayName")
        item_workspace_id = result.get("folderObjectId")
        parent_artifact_id = result.get("parentArtifactObjectId")

        type = "Lakehouse" if item_type == sql else item_type
        item_id = parent_artifact_id if item_type == sql else artifact_id

        sources.append(
            {
                "itemId": item_id,
                "itemName": item_name,
                "itemType": type,
                "workspaceId": item_workspace_id,
                "workspaceName": resolve_workspace_name(item_workspace_id),
                "usesSqlEndpoint": uses_sql_endpoint,
            }
        )

    return sources



import pandas as pd
from uuid import UUID
from typing import Optional
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    resolve_item_id,
    resolve_workspace_id,
    resolve_item_name_and_id,
    resolve_workspace_name_and_id,
)
import sempy_labs._icons as icons


@log
def list_graph_models(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the graph models within a workspace.

    This is a wrapper function for the following API: `Items - List Graph Models <https://learn.microsoft.com/rest/api/fabric/graphmodel/items/list-graph-models>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the graph models within a workspace.
    """

    columns = {
        "Graph Model Name": "string",
        "Graph Model Id": "string",
        "Description": "string",
        "OneLake Root Path": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/GraphModels",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Graph Model Name": v.get("displayName"),
                    "Graph Model Id": v.get("id"),
                    "Description": v.get("description"),
                    "OneLake Root Path": v.get("properties", {}).get("oneLakeRootPath"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def execute_query(
    graph_model: str | UUID, query: str, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Executes a query on the specified graph model.

    This is a wrapper function for the following API: `Items - ExecuteQuery <https://learn.microsoft.com/rest/api/fabric/graphmodel/items/execute-query(preview)>`_.

    Parameters
    ----------
    graph_model : str | uuid.UUID
        The graph model name or ID.
    query : str
        The query string.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    dict
        The response from the API.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=graph_model, type="GraphModel", workspace=workspace_id
    )

    payload = {
        "query": query,
    }
    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/GraphModels/{item_id}/executeQuery?preview=True",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} Executed query on Graph Model '{item_name}' in workspace '{workspace_name}' successfully."
    )

    return response.json()


@log
def get_queryable_graph_type(
    graph_model: str | UUID, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Gets the current queryable graph type.

    This is a wrapper function for the following API: `Items - GetQueryableGraphType <https://learn.microsoft.com/rest/api/fabric/graphmodel/items/get-queryable-graph-type(preview)>`_.

    Parameters
    ----------
    graph_model : str | uuid.UUID
        The graph model name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    dict
        A dictionary showing the current queryable graph type.
    """

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(
        item=graph_model, type="GraphModel", workspace=workspace_id
    )

    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/GraphModels/{item_id}/getQueryableGraphType?preview=True"
    )

    return response.json()



from typing import Optional, List, Tuple
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy.fabric.exceptions import FabricHTTPException
import requests
import pandas as pd
from sempy_labs._authentication import (
    _get_headers,
    ServicePrincipalTokenProvider,
)
from uuid import UUID
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
)
import sempy_labs._authentication as auth


def _add_sll_tag(payload, tags):

    if tags is None:
        payload["tags"] = {"SLL": 1}
    else:
        if "tags" not in payload:
            payload["tags"] = tags
        payload["tags"]["SLL"] = 1

    return payload


@log
def create_fabric_capacity(
    capacity_name: str,
    azure_subscription_id: str,
    resource_group: str,
    region: str,
    sku: str,
    admin_members: str | List[str],
    tags: Optional[dict] = None,
    **kwargs,
):
    """
    This function creates a new Fabric capacity within an Azure subscription.

    This is a wrapper function for the following API: `Fabric Capacities - Create Or Update <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/create-or-update>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group.
    region : str
        The name of the region in which the capacity will be created.
    sku : str
        The `sku size <https://azure.microsoft.com/pricing/details/microsoft-fabric/>`_ of the Fabric capacity.
    admin_members : str | List[str]
        The email address(es) of the admin(s) of the Fabric capacity.
    tags: dict, default=None
        Tag(s) to add to the capacity. Example: {'tagName': 'tagValue'}.
    """

    if isinstance(admin_members, str):
        admin_members = [admin_members]

    # list source: https://learn.microsoft.com/fabric/admin/region-availability
    region_list = [
        "Brazil South",
        "North Europe",
        "UAE North",
        "South Africa North",
        "Australia East",
        "Canada Central",
        "West Europe",
        "Australia Southeast",
        "Canada East",
        "France Central",
        "Central India",
        "East US",
        "Germany West Central",
        "East Asia",
        "East US 2",
        "Norway East",
        "Japan East",
        "North Central US",
        "Sweden Central",
        "Korea Central",
        "South Central US",
        "Switzerland North",
        "Southeast Asia",
        "West US",
        "Switzerland West",
        "South India",
        "West US 2",
        "UK South",
        "West US 3",
        "UK West",
        "brazilsouth",
        "northeurope",
        "uaenorth",
        "southafricanorth",
        "australiaeast",
        "canadacentral",
        "westeurope",
        "australiasoutheast",
        "canadaeast",
        "francecentral",
        "centralindia",
        "eastus",
        "germanywestcentral",
        "eastasia",
        "eastus2",
        "norwayeast",
        "japaneast",
        "northcentralus",
        "swedencentral",
        "koreacentral",
        "southcentralus",
        "switzerlandnorth",
        "southeastasia",
        "westus",
        "switzerlandwest",
        "southindia",
        "westus2",
        "uksouth",
        "westus3",
        "ukwest",
    ]

    valid_regions = [
        region for region in region_list if any(char.isupper() for char in region)
    ]

    if region not in region_list:
        raise ValueError(
            f"{icons.red_dot} Invalid region. Valid options: {valid_regions}."
        )

    token_provider = auth.token_provider.get()
    if token_provider is None:
        token_provider = ServicePrincipalTokenProvider.from_azure_key_vault(
            key_vault_uri=kwargs["key_vault_uri"],
            key_vault_tenant_id=kwargs["key_vault_tenant_id"],
            key_vault_client_id=kwargs["key_vault_client_id"],
            key_vault_client_secret=kwargs["key_vault_client_secret"],
        )
        print(
            f"{icons.info} Please use the 'token_provider' parameter instead of the key vault parameters within this function as the key vault parameters have been deprecated."
        )
    headers = _get_headers(token_provider, audience="azure")

    if resource_group is None:
        dfRG = list_resource_groups(
            azure_subscription_id=azure_subscription_id,
            filter="resourceType eq 'Microsoft.PowerBIDedicated/capacities'",
        )
        dfRG_filt = dfRG[
            dfRG["Resource Group Name"]
            == capacity_name.removesuffix(icons.migrate_capacity_suffix)
        ]
        if not dfRG_filt.empty:
            resource_group = dfRG_filt["Resource Group Name"].iloc[0]
            print(
                f"{icons.yellow_dot} Override resource group flag detected for A SKUs - using the existing resource group '{resource_group}' for the '{capacity_name}' capacity."
            )
    else:
        # Attempt to get the resource group
        try:
            dfRG = get_resource_group(
                azure_subscription_id=azure_subscription_id,
                resource_group=resource_group,
            )
            if dfRG["Location"].iloc[0] != region:
                print(
                    f"{icons.yellow_dot} The '{resource_group}' resource group exists, but in a different region."
                )
        except Exception:
            # If the resource group does not exist, create it
            print(
                f"{icons.yellow_dot} The '{resource_group}' resource group does not exist."
            )
            print(
                f"{icons.in_progress} Creating the '{resource_group}' resource group in the '{region}' region"
            )
            create_or_update_resource_group(
                azure_subscription_id=azure_subscription_id,
                resource_group=resource_group,
                region=region,
            )

    payload = {
        "properties": {"administration": {"members": admin_members}},
        "sku": {"name": sku, "tier": "Fabric"},
        "location": region,
    }

    payload = _add_sll_tag(payload, tags)

    print(
        f"{icons.in_progress} Creating the '{capacity_name}' capacity as an '{sku}' SKU within the '{region}' region..."
    )

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version={icons.azure_api_version}"

    response = requests.put(url, headers=headers, json=payload)

    if response.status_code not in [200, 201]:
        raise FabricHTTPException(response)

    print(
        f"{icons.green_dot} Successfully created the '{capacity_name}' capacity within the '{region}' region."
    )


@log
def list_vcores() -> pd.DataFrame:

    columns = {
        "Total Purchased Cores": "int",
        "Available Cores": "int",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request="capacities/vcores")
    response_json = response.json()
    new_data = {
        "Total Purchased Cores": response_json.get("totalPurchasedCores"),
        "Available Cores": response_json.get("availableCores"),
    }

    df = pd.DataFrame([new_data], columns=columns.keys())
    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def get_capacity_resource_governance(capacity_name: str):

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Display Name"] == capacity_name]
    capacity_id = dfC_filt["Id"].iloc[0].upper()

    response = _base_api(request=f"capacities/{capacity_id}/resourceGovernance")

    return response.json().get("workloadSettings", {})


@log
def suspend_fabric_capacity(
    capacity_name: str,
    azure_subscription_id: str,
    resource_group: str,
):
    """
    This function suspends a Fabric capacity.

    This is a wrapper function for the following API: `Fabric Capacities - Suspend <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/suspend?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Fabric/capacities/{capacity_name}/suspend?api-version={icons.azure_api_version}"

    _base_api(request=url, client="azure", method="post", status_codes=202)

    print(f"{icons.green_dot} The '{capacity_name}' capacity has been suspended.")


@log
def resume_fabric_capacity(
    capacity_name: str,
    azure_subscription_id: str,
    resource_group: str,
):
    """
    This function resumes a Fabric capacity.

    This is a wrapper function for the following API: `Fabric Capacities - Resume <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/resume?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Fabric/capacities/{capacity_name}/resume?api-version={icons.azure_api_version}"

    _base_api(request=url, client="azure", method="post", status_codes=202)

    print(f"{icons.green_dot} The '{capacity_name}' capacity has been resumed.")


@log
def delete_embedded_capacity(
    capacity_name: str,
    azure_subscription_id: str,
    resource_group: str,
):
    """
    This function deletes a Power BI Embedded capacity.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.PowerBIDedicated/capacities/{capacity_name}?api-version={icons.azure_api_version}"

    _base_api(request=url, client="azure", method="delete", status_codes=[200, 202])

    print(f"{icons.green_dot} The '{capacity_name}' capacity has been deleted.")


@log
def delete_premium_capacity(capacity: str | UUID, **kwargs):
    """
    This function deletes a Power BI Premium capacity.

    Parameters
    ----------
    capacity : str | uuid.UUID
        Name or ID of the Fabric capacity.
    """
    from sempy_labs._helper_functions import resolve_capacity_id

    if "capacity_name" in kwargs:
        capacity = kwargs["capacity_name"]
        print(
            f"{icons.warning} The 'capacity_name' parameter is deprecated. Please use 'capacity' instead."
        )

    capacity_id = resolve_capacity_id(capacity=capacity).upper()

    _base_api(request=f"capacities/{capacity_id}", method="delete", status_codes=204)

    print(f"{icons.green_dot} The '{capacity}' capacity has been deleted.")


@log
def delete_fabric_capacity(
    capacity_name: str,
    azure_subscription_id: str,
    resource_group: str,
):
    """
    This function deletes a Fabric capacity.

    This is a wrapper function for the following API: `Fabric Capacities - Delete <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/delete?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version={icons.azure_api_version}"

    _base_api(request=url, client="azure", method="delete", status_codes=202)

    print(f"{icons.green_dot} The '{capacity_name}' capacity has been deleted.")


@log
def update_fabric_capacity(
    capacity_name: str,
    azure_subscription_id: str,
    resource_group: str,
    sku: Optional[str] = None,
    admin_members: Optional[str | List[str]] = None,
    tags: Optional[dict] = None,
):
    """
    This function updates a Fabric capacity's properties.

    This is a wrapper function for the following API: `Fabric Capacities - Update <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/update?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group.
    sku : str, default=None
         The `sku size <https://azure.microsoft.com/pricing/details/microsoft-fabric/>`_ of the Fabric capacity.
    admin_members : str | List[str], default=None
        The email address(es) of the admin(s) of the Fabric capacity.
    tags : dict, default=None
        Tag(s) to add to the capacity. Example: {'tagName': 'tagValue'}.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Fabric/capacities/{capacity_name}?api-version={icons.azure_api_version}"

    get_response = _base_api(request=url, client="azure")

    get_json = get_response.json()
    current_sku = get_json.get("sku", {}).get("name")
    current_admins = (
        get_json.get("properties", {}).get("administration", {}).get("members")
    )
    current_tags = get_json.get("tags")

    payload = {}
    payload["sku"] = {
        "name": current_sku,
        "tier": "Fabric",
    }
    payload["tags"] = current_tags
    payload["properties"] = get_json["properties"]

    if sku is not None:
        payload["sku"]["name"] = sku
    if admin_members is not None:
        payload["properties"] = {"administration": {"members": admin_members}}
    if tags is not None:
        payload["tags"] = tags

    # Do not proceed if no properties are being changed
    if current_sku == sku and current_admins == admin_members and current_tags == tags:
        print(
            f"{icons.yellow_dot} The properties of the '{capacity_name}' are the same as those specified in the parameters of this function. No changes have been made."
        )
        return

    payload = _add_sll_tag(payload, tags)
    _base_api(
        request=url, client="azure", method="patch", payload=payload, status_codes=202
    )

    print(
        f"{icons.green_dot} The '{capacity_name}' capacity has been updated accordingly."
    )


@log
def check_fabric_capacity_name_availablility(
    capacity_name: str,
    azure_subscription_id: str,
    region: str,
    **kwargs,
) -> bool:
    """
    This function updates a Fabric capacity's properties.

    This is a wrapper function for the following API: `Fabric Capacities - Check Name Availability <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/check-name-availability?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity_name : str
        Name of the Fabric capacity.
    azure_subscription_id : str
        The Azure subscription ID.
    region : str
        The region name.

    Returns
    -------
    bool
        An indication as to whether the Fabric capacity name is available or not.
    """

    payload = {"name": capacity_name, "type": "Microsoft.Fabric/capacities"}

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/providers/Microsoft.Fabric/locations/{region}/checkNameAvailability?api-version={icons.azure_api_version}"

    response = _base_api(
        request=url, client="azure", method="post", payload=payload, status_codes=202
    )

    return bool(response.json().get("nameAvailable"))


@log
def create_resource_group(
    azure_subscription_id: str,
    resource_group: str,
    region: str,
    **kwargs,
):
    """
    This function creates a resource group in a region within an Azure subscription.

    This is a wrapper function for the following API: `Resource Groups - Create Or Update <https://learn.microsoft.com/rest/api/resources/resource-groups/create-or-update>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the Azure resource group to be created.
    region : str
        The name of the region in which the resource group will be created.
    """

    if check_resource_group_existence(
        azure_subscription_id=azure_subscription_id,
        resource_group=resource_group,
    ):
        print(
            f"{icons.info} The '{resource_group}' resource group already exists in the '{region}' region within the '{azure_subscription_id}' Azure subscription."
        )
        return

    create_or_update_resource_group(
        azure_subscription_id=azure_subscription_id,
        resource_group=resource_group,
        region=region,
    )


@log
def list_skus_for_capacity(
    capacity: str,
    azure_subscription_id: str,
    resource_group: str,
) -> pd.DataFrame:
    """
    Lists eligible SKUs for a Microsoft Fabric resource.

    This is a wrapper function for the following API: `Fabric Capacities - List Skus For Capacity <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/list-skus-for-capacity?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str
        The capacity name.
    azure_subscription_id : str
        The Azure subscription ID.
    resource_group : str
        The name of the resource group.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of eligible SKUs for a Microsoft Fabric resource.
    """

    columns = {
        "Resource Type": "string",
        "Sku": "string",
        "Sku Tier": "string",
    }
    df = _create_dataframe(columns=columns)

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Fabric/capacities/{capacity}/skus?api-version=2023-11-01"
    response = _base_api(request=url, client="azure")

    rows = []
    for v in response.json().get("value", []):
        sku = v.get("sku", {})
        rows.append(
            {
                "Resource Type": v.get("resourceType"),
                "Sku": sku.get("name"),
                "Sku Tier": sku.get("tier"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def list_skus(
    azure_subscription_id: str,
) -> pd.DataFrame:
    """
    Lists eligible SKUs for Microsoft Fabric resource provider.

    This is a wrapper function for the following API: `Fabric Capacities - List Skus For Capacity <https://learn.microsoft.com/rest/api/microsoftfabric/fabric-capacities/list-skus?view=rest-microsoftfabric-2023-11-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of eligible SKUs for Microsoft Fabric resource provider.
    """

    columns = {
        "Sku": "string",
        "Locations": "str",
    }
    df = _create_dataframe(columns=columns)

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/providers/Microsoft.Fabric/skus?api-version=2023-11-01"
    response = _base_api(request=url, client="azure")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Sku": v.get("name"),
                "Locations": v.get("locations", []),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def list_subscriptions() -> pd.DataFrame:
    """
    Gets all subscriptions for a tenant.

    This is a wrapper function for the following API: `Subscriptions - List <https://learn.microsoft.com/rest/api/resources/subscriptions/list?view=rest-resources-2022-12-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all subscriptions for a tenant.
    """

    columns = {
        "Subscription Id": "string",
        "Subscription Name": "string",
        "Tenant Id": "string",
        "State": "string",
        "Location Placement Id": "string",
        "Quota Id": "string",
        "Spending Limit": "string",
        "Authorization Source": "string",
        "Managed by Tenants": "string",
        "Tags": "string",
    }
    df = _create_dataframe(columns=columns)

    url = "https://management.azure.com/subscriptions?api-version=2022-12-01"
    response = _base_api(request=url, client="azure")

    rows = []
    for v in response.json().get("value", []):
        policy = v.get("subscriptionPolicies", {})
        tenants = v.get("managedByTenants")
        rows.append(
            {
                "Subscription Id": v.get("subscriptionId"),
                "Subscription Name": v.get("displayName"),
                "Tenant Id": v.get("tenantId"),
                "State": v.get("state"),
                "Location Placement Id": policy.get("locationPlacementId"),
                "Quota Id": policy.get("quotaId"),
                "Spending Limit": policy.get("spendingLimit"),
                "Authorization Source": v.get("authorizationSource"),
                "Managed by Tenants": tenants if tenants is not None else [],
                "Tags": v.get("tags", {}),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def get_subscription(azure_subscription_id: str) -> pd.DataFrame:
    """
    Gets details about a specified subscription.

    This is a wrapper function for the following API: `Subscriptions - Get <https://learn.microsoft.com/rest/api/resources/subscriptions/get?view=rest-resources-2022-12-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing details of a specific subscription.
    """

    columns = {
        "Subscription Id": "string",
        "Subscription Name": "string",
        "Tenant Id": "string",
        "State": "string",
        "Location Placement Id": "string",
        "Quota Id": "string",
        "Spending Limit": "string",
        "Authorization Source": "string",
        "Managed by Tenants": "string",
        "Tags": "string",
    }
    df = _create_dataframe(columns=columns)

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}?api-version=2022-12-01"

    response = _base_api(request=url, client="azure")
    v = response.json()
    policy = v.get("subscriptionPolicies", {})
    tenants = v.get("managedByTenants")
    new_data = {
        "Subscription Id": v.get("subscriptionId"),
        "Subscription Name": v.get("displayName"),
        "Tenant Id": v.get("tenantId"),
        "State": v.get("state"),
        "Location Placement Id": policy.get("locationPlacementId"),
        "Quota Id": policy.get("quotaId"),
        "Spending Limit": policy.get("spendingLimit"),
        "Authorization Source": v.get("authorizationSource"),
        "Managed by Tenants": tenants if tenants is not None else [],
        "Tags": v.get("tags", {}),
    }

    df = pd.DataFrame([new_data], columns=columns.keys())

    return df


def _resolve_subscription_name_and_id(
    azure_subscription: str | UUID,
) -> Tuple[str, UUID]:

    if _is_valid_uuid(azure_subscription):
        subscription_id = azure_subscription
        df = get_subscription(azure_subscription_id=subscription_id)
        if df.empty:
            raise ValueError(f"{icons.red_dot} The subscription ID does not exist.")
        subscription_name = df["Subscription Name"].iloc[0]
    else:
        subscription_name = azure_subscription
        df = list_subscriptions()
        df_filt = df[df["Subscription Name"] == subscription_name]
        if df_filt.empty:
            raise ValueError(f"{icons.red_dot} The subscription name does not exist.")
        subscription_id = df_filt["Subscription Id"].iloc[0]

    return subscription_name, subscription_id


@log
def list_tenants() -> pd.DataFrame:
    """
    Gets the tenants for your account.

    This is a wrapper function for the following API: `Tenants - List <https://learn.microsoft.com/rest/api/resources/tenants/list?view=rest-resources-2022-12-01>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all tenants for your account.
    """

    columns = {
        "Tenant Id": "string",
        "Tenant Name": "string",
        "Country Code": "string",
        "Domains": "string",
        "Tenant Category": "string",
        "Default Domain": "string",
        "Tenant Type": "string",
        "Tenant Branding Logo Url": "string",
    }
    df = _create_dataframe(columns=columns)

    url = "https://management.azure.com/tenants?api-version=2022-12-01"

    response = _base_api(request=url, client="azure")

    rows = []
    for v in response.json().get("value", []):
        d = v.get("domains")
        rows.append(
            {
                "Tenant Id": v.get("tenantId"),
                "Tenant Name": v.get("displayName"),
                "Country Code": v.get("countryCode"),
                "Domains": d if d is not None else "",
                "Tenant Category": v.get("tenantCategory"),
                "Default Domain": v.get("defaultDomain"),
                "Tenant Type": v.get("tenantType"),
                "Tenant Branding Logo Url": v.get("tenantBrandingLogoUrl"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def create_or_update_resource_group(
    azure_subscription_id: str,
    resource_group: str,
    region: str,
):
    """
    Creates or updates a resource group.

    This is a wrapper function for the following API: `Resource Groups - Create Or Update <https://learn.microsoft.com/rest/api/resources/resource-groups/create-or-update>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription Id.
    resource_group : str
        The name of the resource group.
    region : str
        The name of the region.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourcegroups/{resource_group}?api-version=2021-04-01"

    payload = {
        "location": region,
    }

    _base_api(
        request=url,
        client="azure",
        method="put",
        payload=payload,
        status_codes=[200, 201],
    )

    print(
        f"{icons.green_dot} The '{resource_group}' resource group has been created/updated."
    )


@log
def create_storage_account(
    azure_subscription_id: str,
    resource_group: str,
    storage_account: str,
    region: str,
):
    """
    Asynchronously creates a new storage account with the specified parameters. If an account is already created and a subsequent create request is issued with different properties, the account properties will be updated. If an account is already created and a subsequent create or update request is issued with the exact same set of properties, the request will succeed.

    This is a wrapper function for the following API: `Storage Accounts - Create <https://learn.microsoft.com/rest/api/storagerp/storage-accounts/create>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription Id.
    resource_group : str
        The name of the resource group.
    storage_account : str
        The name of the storage account to be created.
    region : str
        The name of the region.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}/providers/Microsoft.Storage/storageAccounts/{storage_account}?api-version=2018-02-01"

    payload = {
        "sku": {"name": "Standard_GRS"},
        "kind": "StorageV2",
        "location": region,
    }

    _base_api(request=url, client="azure", method="put", payload=payload)

    print(
        f"{icons.green_dot} The '{storage_account}' storage account has been created."
    )


@log
def list_storage_accounts(
    azure_subscription_id: str,
    resource_group: Optional[str] = None,
) -> pd.DataFrame:
    """
    Lists all the storage accounts available under the subscription (or resource group). Note that storage keys are not returned; use the ListKeys operation for this.

    This is a wrapper function for the following APIs: `Storage Accounts - List <https://learn.microsoft.com/rest/api/storagerp/storage-accounts/list>`_, `Storage Accounts - List By Resource Group <https://learn.microsoft.com/rest/api/storagerp/storage-accounts/list-by-resource-group>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription Id.
    resource_group : str, default=None
        If set to None, retrieves all storage accounts for the subscription. If not None, shows the storage accounts within that resource group.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all storage accounts within the subscription (or resource group).
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}"

    if resource_group is not None:
        url += f"/resourceGroups/{resource_group}"

    url += "/providers/Microsoft.Storage/storageAccounts?api-version=2023-05-01"

    columns = {
        "Storage Account Id": "string",
        "Storage Account Name": "string",
        "Kind": "string",
        "Location": "string",
        "Sku Name": "string",
        "Sku Tier": "string",
        "Is HNS Enabled": "bool",
        "Creation Time": "datetime",
        "Web Endpoint": "string",
        "DFS Endpoint": "string",
        "Blob Endpoint": "string",
        "File Endpoint": "string",
        "Queue Endpoint": "string",
        "Table Endpoint": "string",
        "Primary Location": "string",
        "Provisioning State": "string",
        "Secondary Location": "string",
        "Status of Primary": "string",
        "Status of Secondary": "string",
        "Supports HTTPS Traffic Only": "bool",
        "Tags": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request=url, client="azure")

    rows = []
    for v in response.json().get("value", []):
        p = v.get("properties", {})
        rows.append(
            {
                "Storage Account Id": v.get("id"),
                "Storage Account Name": v.get("name"),
                "Kind": v.get("kind"),
                "Location": v.get("location"),
                "Sku Name": v.get("sku", {}).get("name"),
                "Sku Tier": v.get("sku", {}).get("tier"),
                "Is HNS Enabled": p.get("isHnsEnabled"),
                "Creation Time": p.get("creationTime"),
                "Web Endpoint": p.get("primaryEndpoints", {}).get("web"),
                "DFS Endpoint": p.get("primaryEndpoints", {}).get("dfs"),
                "Blob Endpoint": p.get("primaryEndpoints", {}).get("blob"),
                "File Endpoint": p.get("primaryEndpoints", {}).get("file"),
                "Queue Endpoint": p.get("primaryEndpoints", {}).get("queue"),
                "Table Endpoint": p.get("primaryEndpoints", {}).get("table"),
                "Primary Location": p.get("primaryLocation"),
                "Provisioning State": p.get("provisioningState"),
                "Secondary Location": p.get("secondaryLocation"),
                "Status of Primary": p.get("statusOfPrimary"),
                "Status of Secondary": p.get("statusOfSecondary"),
                "Supports HTTPS Traffic Only": p.get("supportsHttpsTrafficOnly"),
                "Tags": v.get("tags"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def check_resource_group_existence(
    azure_subscription_id: str, resource_group: str
) -> bool:
    """
    Checks whether a resource group exists.

    This is a wrapper function for the following API: `Resource Groups - Check Existence <https://learn.microsoft.com/rest/api/resources/resource-groups/check-existence>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription Id.
    resource_group : str
        The name of the resource group.

    Returns
    -------
    bool
        True/False indicating if the resource group exists or not.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}?api-version=2021-04-01"

    response = _base_api(request=url, client="azure", status_codes=[200, 204, 404])

    if response.status_code == 200:
        return True
    elif response.status_code in [204, 404]:
        return False


@log
def list_resource_groups(
    azure_subscription_id: str,
    filter: Optional[str] = None,
    top: Optional[int] = None,
) -> pd.DataFrame:
    """
    Lists all resource groups within a subscription.

    This is a wrapper function for the following API: `Resource Groups - List <https://learn.microsoft.com/rest/api/resources/resource-groups/list>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription Id.
    filter : str, default=None
        The filter to apply to the operation. Example: filter="tagname eq 'tagvalue'".
    top : int, default=None
        The number of results to return. If not specified, returns all results.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all resource groups within the subscription.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups?"

    if filter is not None:
        url += f"$filter={filter}&"
    if top is not None:
        url += f"$top={top}&"

    url += "api-version=2021-04-01"

    columns = {
        "Resource Group Name": "string",
        "Location": "string",
        "Tags": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request=url, client="azure")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Resource Group Id": v.get("id"),
                "Resource Group Name": v.get("name"),
                "Location": v.get("location"),
                "Managed By": v.get("managedBy"),
                "Tags": v.get("tags"),
                "Type": v.get("type"),
                "Provisioning State": v.get("properties", {}).get("provisioningState"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def get_resource_group(azure_subscription_id: str, resource_group: str) -> pd.DataFrame:
    """
    Gets details about a specified resource group.

    This is a wrapper function for the following API: `Resource Groups - Get <https://learn.microsoft.com/rest/api/resources/resource-groups/get>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription Id.
    resource_group : str
        The name of the resource group.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing details of a specific resource group.
    """

    url = f"https://management.azure.com/subscriptions/{azure_subscription_id}/resourceGroups/{resource_group}?api-version=2021-04-01"
    response = _base_api(request=url, client="azure")

    v = response.json()
    new_data = {
        "Resource Group Id": v.get("id"),
        "Resource Group Name": v.get("name"),
        "Location": v.get("location"),
        "Managed By": v.get("managedBy"),
        "Tags": v.get("tags"),
        "Type": v.get("type"),
        "Provisioning State": v.get("properties", {}).get("provisioningState"),
    }

    return pd.DataFrame(new_data, index=[0])


def list_capacities() -> pd.DataFrame:
    """
    Shows the capacities and their properties.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the capacities and their properties
    """

    columns = {
        "Id": "string",
        "Display Name": "string",
        "Sku": "string",
        "Region": "string",
        "State": "string",
        "Admins": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request="/v1.0/myorg/capacities", client="fabric_sp")

    rows = []
    for i in response.json().get("value", []):
        rows.append(
            {
                "Id": i.get("id").lower(),
                "Display Name": i.get("displayName"),
                "Sku": i.get("sku"),
                "Region": i.get("region"),
                "State": i.get("state"),
                "Admins": [i.get("admins", [])],
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df



from ._items import (
    list_environments,
    create_environment,
    delete_environment,
    publish_environment,
)
from ._pubstage import (
    get_published_spark_compute,
    get_staging_spark_compute,
    list_published_libraries,
    list_staging_libraries,
)

__all__ = [
    "list_environments",
    "create_environment",
    "delete_environment",
    "publish_environment",
    "get_published_spark_compute",
    "get_staging_spark_compute",
    "list_published_libraries",
    "list_staging_libraries",
]



import pandas as pd
from sempy_labs._helper_functions import (
    _create_dataframe,
    _update_dataframe_datatypes,
    _base_api,
)
from sempy._utils._log import log
from uuid import UUID
import sempy_labs._icons as icons


@log
def list_tenant_keys() -> pd.DataFrame:
    """
    Returns the encryption keys for the tenant.

    This is a wrapper function for the following API: `Admin - Get Power BI Encryption Keys <https://learn.microsoft.com/rest/api/power-bi/admin/get-power-bi-encryption-keys>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the encryption keys for the tenant.
    """

    columns = {
        "Tenant Key Id": "str",
        "Tenant Key Name": "str",
        "Key Vault Key Identifier": "str",
        "Is Default": "bool",
        "Created At": "datetime",
        "Updated At": "datetime",
    }

    df = _create_dataframe(columns=columns)

    result = _base_api(
        request="/v1.0/myorg/admin/tenantKeys", client="fabric_sp"
    ).json()

    rows = []
    for i in result.get("value", []):
        rows.append(
            {
                "Tenant Key Id": i.get("id"),
                "Tenant Key Name": i.get("name"),
                "Key Vault Key Identifier": i.get("keyVaultKeyIdentifier"),
                "Is Default": i.get("isDefault"),
                "Created At": i.get("createdAt"),
                "Updated At": i.get("updatedAt"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def rotate_tenant_key(tenant_key_id: UUID, key_vault_key_identifier: str):
    """
    Rotate the encryption key for Power BI workspaces assigned to a capacity.

    This is a wrapper function for the following API: `Admin - Rotate Power BI Encryption Key <https://learn.microsoft.com/rest/api/power-bi/admin/rotate-power-bi-encryption-key>`_.

    Parameters
    ----------
    tenant_key_id : uuid.UUID
        The tenant key ID
    key_vault_key_identifier : str
        The URI that uniquely specifies the encryption key in Azure Key Vault
    """

    payload = {
        "keyVaultKeyIdentifier": key_vault_key_identifier,
    }

    _base_api(
        request=f"/v1.0/myorg/admin/tenantKeys/{tenant_key_id}/Default.Rotate",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The encryption key for tenant {tenant_key_id} has been rotated successfully using the '{key_vault_key_identifier}' Key Vault key identifier."
    )



from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    delete_item,
    resolve_workspace_id,
)
import pandas as pd
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
from sempy._utils._log import log


@log
def create_warehouse(
    warehouse: str,
    description: Optional[str] = None,
    case_insensitive_collation: bool = False,
    workspace: Optional[str | UUID] = None,
) -> UUID:
    """
    Creates a Fabric warehouse.

    This is a wrapper function for the following API: `Items - Create Warehouse <https://learn.microsoft.com/rest/api/fabric/warehouse/items/create-warehouse>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    warehouse: str
        Name of the warehouse.
    description : str, default=None
        A description of the warehouse.
    case_insensitive_collation: bool, default=False
        If True, creates the warehouse with case-insensitive collation.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    uuid.UUID
        The ID of the created warehouse.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = {"displayName": warehouse}

    if description:
        payload["description"] = description
    if case_insensitive_collation:
        payload.setdefault("creationPayload", {})
        payload["creationPayload"][
            "defaultCollation"
        ] = "Latin1_General_100_CI_AS_KS_WS_SC_UTF8"

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses",
        payload=payload,
        method="post",
        lro_return_json=True,
        status_codes=[201, 202],
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{warehouse}' warehouse has been created within the '{workspace_name}' workspace."
    )

    return result.get("id")


@log
def list_warehouses(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the warehouses within a workspace.

    This is a wrapper function for the following API: `Items - List Warehouses <https://learn.microsoft.com/rest/api/fabric/warehouse/items/list-warehouses>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the warehouses within a workspace.
    """

    columns = {
        "Warehouse Name": "string",
        "Warehouse Id": "string",
        "Description": "string",
        "Connection Info": "string",
        "Created Date": "datetime",
        "Last Updated Time": "datetime",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})

            rows.append(
                {
                    "Warehouse Name": v.get("displayName"),
                    "Warehouse Id": v.get("id"),
                    "Description": v.get("description"),
                    "Connection Info": prop.get("connectionInfo"),
                    "Created Date": prop.get("createdDate"),
                    "Last Updated Time": prop.get("lastUpdatedTime"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def delete_warehouse(name: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a Fabric warehouse.

    This is a wrapper function for the following API: `Items - Delete Warehouse <https://learn.microsoft.com/rest/api/fabric/warehouse/items/delete-warehouse>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str | uuid.UUID
        Name or ID of the warehouse.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=name, type="Warehouse", workspace=workspace)


@log
def get_warehouse_tables(
    warehouse: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of the tables in the Fabric warehouse. This function is based on INFORMATION_SCHEMA.TABLES.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        Name or ID of the Fabric warehouse.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the tables in the Fabric warehouse.
    """

    from sempy_labs._sql import ConnectWarehouse

    with ConnectWarehouse(warehouse=warehouse, workspace=workspace) as sql:
        df = sql.query(
            """
        SELECT TABLE_SCHEMA AS [Schema], TABLE_NAME AS [Table Name], TABLE_TYPE AS [Table Type]
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_TYPE = 'BASE TABLE'
        """
        )

    return df


@log
def get_warehouse_columns(
    warehouse: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of the columns in each table within the Fabric warehouse. This function is based on INFORMATION_SCHEMA.COLUMNS.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        Name or ID of the Fabric warehouse.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the columns in each table within the Fabric warehouse.
    """

    from sempy_labs._sql import ConnectWarehouse

    with ConnectWarehouse(warehouse=warehouse, workspace=workspace) as sql:
        df = sql.query(
            """
        SELECT t.TABLE_SCHEMA AS [Schema], t.TABLE_NAME AS [Table Name], c.COLUMN_NAME AS [Column Name], c.DATA_TYPE AS [Data Type], c.IS_NULLABLE AS [Is Nullable], c.CHARACTER_MAXIMUM_LENGTH AS [Character Max Length]
        FROM INFORMATION_SCHEMA.TABLES AS t
        LEFT JOIN INFORMATION_SCHEMA.COLUMNS AS c
        ON t.TABLE_NAME = c.TABLE_NAME
        AND t.TABLE_SCHEMA = c.TABLE_SCHEMA
        WHERE t.TABLE_TYPE = 'BASE TABLE'
        """
        )

    return df



from typing import Dict, Literal, Optional
from azure.core.credentials import AccessToken, TokenCredential
from azure.identity import ClientSecretCredential
from sempy._utils._log import log
from contextlib import contextmanager
import contextvars


class ServicePrincipalTokenProvider(TokenCredential):
    """
    A class to acquire authentication token with Service Principal.

    For more information on Service Principal see: `Application and service principal objects in Microsoft Entra ID <https://learn.microsoft.com/en-us/entra/identity-platform/app-objects-and-service-principals?tabs=browser#service-principal-object>`_
    """

    _shorthand_scopes: Dict[str, str] = {
        "pbi": "https://analysis.windows.net/powerbi/api/.default",
        "storage": "https://storage.azure.com/.default",
        "azure": "https://management.azure.com/.default",
        "graph": "https://graph.microsoft.com/.default",
        "asazure": "https://{region}.asazure.windows.net/.default",
        "keyvault": "https://vault.azure.net/.default",
    }

    def __init__(self, credential: ClientSecretCredential):

        self.credential = credential

    @classmethod
    def from_aad_application_key_authentication(
        cls, tenant_id: str, client_id: str, client_secret: str
    ) -> "ServicePrincipalTokenProvider":
        """
        Generates the ServicePrincipalTokenProvider, providing the Service Principal information.

        ***USE THIS ONE ONLY FOR TEST PURPOSE. FOR PRODUCTION WE RECOMMEND CALLING ServicePrincipalTokenProvider.from_azure_key_vault()***

        Parameters
        ----------
        tenant_id : str
            The Fabric Tenant ID.
        client_id : str
            The Service Principal Application Client ID.
        client_secret : str
            The Service Principal Client Secret.

        Returns
        -------
        ServicePrincipalTokenProvider
            Token provider to be used with FabricRestClient or PowerBIRestClient.
        """
        credential = ClientSecretCredential(
            tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
        )

        cls.tenant_id = tenant_id
        cls.client_id = client_id
        cls.client_secret = client_secret

        return cls(credential)

    @classmethod
    def from_azure_key_vault(
        cls,
        key_vault_uri: str,
        key_vault_tenant_id: str,
        key_vault_client_id: str,
        key_vault_client_secret: str,
    ) -> "ServicePrincipalTokenProvider":
        """
        Generates the ServicePrincipalTokenProvider, providing the Azure Key Vault details.

        For more information on Azure Key Vault, `click here <https://learn.microsoft.com/en-us/azure/key-vault/general/overview>`_.

        Parameters
        ----------
        key_vault_uri : str
            Azure Key Vault URI.
        key_vault_tenant_id : str
            Name of the secret in the Key Vault with the Fabric Tenant ID.
        key_vault_client_id : str
            Name of the secret in the Key Vault with the Service Principal Client ID.
        key_vault_client_secret : str
            Name of the secret in the Key Vault with the Service Principal Client Secret.

        Returns
        -------
        ServicePrincipalTokenProvider
            Token provider to be used with FabricRestClient or PowerBIRestClient.
        """

        import notebookutils

        tenant_id = notebookutils.credentials.getSecret(
            key_vault_uri, key_vault_tenant_id
        )
        client_id = notebookutils.credentials.getSecret(
            key_vault_uri, key_vault_client_id
        )
        client_secret = notebookutils.credentials.getSecret(
            key_vault_uri, key_vault_client_secret
        )

        credential = ClientSecretCredential(
            tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
        )

        cls.tenant_id = tenant_id
        cls.client_id = client_id
        cls.client_secret = client_secret

        return cls(credential)

    def __call__(
        self,
        audience: Literal[
            "pbi", "storage", "azure", "graph", "asazure", "keyvault"
        ] = "pbi",
        region: Optional[str] = None,
    ) -> str:
        """
        Parameters
        ----------
        audience : typing.Literal["pbi", "storage", "azure", "graph", "asazure", "keyvault"] = "pbi") -> str
            Literal if it's for PBI/Fabric API call or OneLake/Storage Account call.
        region : str, default=None
            The region of the Azure Analysis Services. For example: 'westus2'.
        """
        # Check if audience is supported
        if audience not in self._shorthand_scopes:
            raise NotImplementedError

        return self.get_token(audience, region=region).token

    def get_token(self, *scopes, **kwargs) -> AccessToken:
        """
        Gets a token for the specified scopes.

        Parameters
        ----------
        *scopes : str
            The scopes for which to obtain a token.
        **kwargs : dict
            Additional parameters to pass to the token request.

        Returns
        -------
        AccessToken
            The access token.
        """
        if len(scopes) == 0:
            scopes = ("pbi",)

        region = kwargs.pop("region", None)
        scopes = [
            self._get_fully_qualified_scope(scope, region=region) for scope in scopes
        ]
        return self.credential.get_token(*scopes, **kwargs)

    def _get_fully_qualified_scope(
        self, scope: str, region: Optional[str] = None
    ) -> str:
        """
        Resolve to fully qualified scope if Fabric short-handed scope is given.
        Otherwise, return the original scope.

        Parameters
        ----------
        scope : str
            The scope to resolve.
        region : str, default=None
            The specific region to use to resolve scope.
            Required if scope is "asazure".

        Returns
        -------
        str
            The resolved scope.
        """
        fully_qualified_scope = self._shorthand_scopes.get(scope, scope)

        if scope == "asazure":
            if region is None:
                raise ValueError("Region is required for 'asazure' scope")
            return fully_qualified_scope.format(region=region)

        return fully_qualified_scope


def _get_headers(
    token_provider: TokenCredential,
    audience: Literal[
        "pbi", "storage", "azure", "graph", "asazure", "keyvault"
    ] = "azure",
):
    """
    Generates headers for an API request.
    """

    token = token_provider.get_token(audience).token

    headers = {"Authorization": f"Bearer {token}"}

    if audience == "graph":
        headers["ConsistencyLevel"] = "eventual"
    else:
        headers["Content-Type"] = "application/json"

    return headers


token_provider = contextvars.ContextVar("token_provider", default=None)


@log
@contextmanager
def service_principal_authentication(
    key_vault_uri: str,
    key_vault_tenant_id: str,
    key_vault_client_id: str,
    key_vault_client_secret: str,
):
    """
    Establishes an authentication via Service Principal.

    Parameters
    ----------
    key_vault_uri : str
        Azure Key Vault URI.
    key_vault_tenant_id : str
        Name of the secret in the Key Vault with the Fabric Tenant ID.
    key_vault_client_id : str
        Name of the secret in the Key Vault with the Service Principal Client ID.
    key_vault_client_secret : str
        Name of the secret in the Key Vault with the Service Principal Client Secret.
    """

    # Save the prior state
    prior_token = token_provider.get()

    # Set the new token_provider in a thread-safe manner
    token_provider.set(
        ServicePrincipalTokenProvider.from_azure_key_vault(
            key_vault_uri=key_vault_uri,
            key_vault_tenant_id=key_vault_tenant_id,
            key_vault_client_id=key_vault_client_id,
            key_vault_client_secret=key_vault_client_secret,
        )
    )
    try:
        from sempy.fabric import set_service_principal

        with set_service_principal(
            (key_vault_uri, key_vault_tenant_id),
            (key_vault_uri, key_vault_client_id),
            client_secret=(key_vault_uri, key_vault_client_secret),
        ):
            yield
    finally:
        # Restore the prior state
        if prior_token is None:
            token_provider.set(None)
        else:
            token_provider.set(prior_token)



from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    delete_item,
    resolve_workspace_id,
)
import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log


@log
def list_snowflake_databases(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the snowflake databases within a workspace.

    This is a wrapper function for the following API: `Items - List Snowflake Databases <https://learn.microsoft.com/rest/api/fabric/snowflakedatabase/items/list-snowflake-databases>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the snowflake databases within a workspace.
    """

    columns = {
        "Snowflake Database Display Name": "string",
        "Snowflake Database Id": "string",
        "Description": "string",
        "Snowflake Database Name": "string",
        "OneLake Tables Path": "string",
        "Snowflake Volume Path": "string",
        "SQL Endpoint Connection String": "string",
        "SQL Endpoint Id": "string",
        "SQL Endpoint Provisioning Status": "string",
        "Default Schema": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/snowflakeDatabases",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})

            rows.append(
                {
                    "Snowflake Database Display Name": v.get("displayName"),
                    "Snowflake Database Id": v.get("id"),
                    "Description": v.get("description"),
                    "Snowflake Database Name": prop.get("snowflakeDatabaseName"),
                    "OneLake Tables Path": prop.get("onelakeTablesPath"),
                    "Snowflake Volume Path": prop.get("snowflakeVolumePath"),
                    "SQL Endpoint Connection String": prop.get(
                        "sqlEndpointProperties", {}
                    ).get("connectionString"),
                    "SQL Endpoint Id": prop.get("sqlEndpointProperties", {}).get("id"),
                    "SQL Endpoint Provisioning Status": prop.get(
                        "sqlEndpointProperties", {}
                    ).get("provisioningStatus"),
                    "Default Schema": prop.get("defaultSchema"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def delete_snowflake_database(
    snowflake_database: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes a Fabric snowflake database.

    This is a wrapper function for the following API: `Items - Delete Snowflake Database <https://learn.microsoft.com/rest/api/fabric/snowflakedatabase/items/delete-snowflake-database>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).
    """

    delete_item(
        item=snowflake_database,
        type="SnowflakeDatabase",
        workspace=workspace,
    )



import os
import base64
import json
import sempy.fabric as fabric
import sempy_labs._icons as icons
from sempy_labs.report._generate_report import get_report_definition
from sempy_labs._generate_semantic_model import get_semantic_model_definition
from sempy_labs._helper_functions import (
    _mount,
    resolve_workspace_name_and_id,
    resolve_item_name,
    resolve_workspace_name,
    resolve_item_name_and_id,
)
from uuid import UUID
from sempy._utils._log import log
from typing import Optional


@log
def save_report_as_pbip(
    report: str | UUID,
    workspace: Optional[str | UUID] = None,
    thick_report: bool = True,
    live_connect: bool = True,
    lakehouse: Optional[str | UUID] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
):
    """
    Saves a report as a .pbip file to the default lakehouse attached to the notebook.

    Parameters
    ----------
    report : str | uuid.UUID
        Name or ID of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    thick_report : bool, default=True
        If set to True, saves the report and underlying semantic model.
        If set to False, saves just the report.
    live_connect : bool, default=True
        If set to True, saves a .pbip live-connected to the workspace in the Power BI / Fabric service.
        If set to False, saves a .pbip with a local model, independent from the Power BI / Fabric service.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID. This will be the lakehouse to which the report is saved.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (report_workspace_name, report_workspace_id) = resolve_workspace_name_and_id(
        workspace
    )
    (report_name, report_id) = resolve_item_name_and_id(
        item=report, type="Report", workspace=workspace
    )
    indent = 2

    local_path = _mount(lakehouse=lakehouse, workspace=lakehouse_workspace)
    save_location = f"{local_path}/Files"

    # Find semantic model info
    dfR = fabric.list_reports(workspace=workspace)
    dfR_filt = dfR[dfR["Id"] == report_id]
    if dfR_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{report} report does not exist within the '{report_workspace_name} workspace."
        )

    dataset_id = dfR_filt["Dataset Id"].iloc[0]
    dataset_workspace_id = dfR_filt["Dataset Workspace Id"].iloc[0]
    dataset_name = resolve_item_name(item_id=dataset_id, workspace=dataset_workspace_id)
    dataset_workspace_name = resolve_workspace_name(dataset_workspace_id)
    path_prefix = f"{save_location}/{report_workspace_name}/{report_name}/{report_name}"

    # Local model not supported if the report and model are in different workspaces
    if dataset_workspace_name != report_workspace_name and not live_connect:
        live_connect = True
        print(
            f"{icons.warning} The '{report_name}' report from the '{report_workspace_name}' workspace is being saved as a live-connected report/model."
        )

    def add_files(name, type, object_workspace):

        path_prefix_full = f"{path_prefix}.{type}"

        if type == "Report":
            dataframe = get_report_definition(report=name, workspace=workspace)
        elif type == "SemanticModel":
            dataframe = get_semantic_model_definition(
                dataset=name, workspace=object_workspace
            )
        else:
            raise NotImplementedError

        # Create and save files based on dataset/report definition
        for _, r in dataframe.iterrows():
            path = r["path"]
            file_content = base64.b64decode(r["payload"])
            file_path = f"{path_prefix_full}/{path}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

            # Update the definition.pbir file for local models
            if not live_connect and type == "Report" and path == "definition.pbir":
                file_content = {
                    "version": "1.0",
                    "datasetReference": {
                        "byPath": {"path": f"../{report_name}.SemanticModel"},
                        "byConnection": None,
                    },
                }

                with open(file_path, "w") as f:
                    json.dump(file_content, f, indent=indent)
            else:
                with open(file_path, "wb") as f:
                    f.write(file_content)

        # Create and save .pbip file for report, converting the file extension
        if type == "Report":
            # Standard .pbip file content
            pbip = {
                "version": "1.0",
                "artifacts": [{"report": {"path": f"{report_name}.Report"}}],
                "settings": {"enableAutoRecovery": True},
            }
            # Ensure the directory exists
            os.makedirs(os.path.dirname(path_prefix), exist_ok=True)
            # Write the .pbip file directly
            pbip_final = f"{path_prefix}.pbip"
            with open(pbip_final, "w") as file:
                json.dump(pbip, file, indent=indent)

    add_files(name=report_name, type="Report", object_workspace=workspace)
    if thick_report:
        add_files(
            name=dataset_name,
            type="SemanticModel",
            object_workspace=dataset_workspace_name,
        )
    print(
        f"{icons.green_dot} The '{report_name}' report within the '{report_workspace_name}' workspace has been saved to this location: {save_location}."
    )



import sempy.fabric as fabric
import pandas as pd
import re
from ..tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID
from .._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    save_as_delta_table,
    retry,
)


@log
def refresh_calc_tables(dataset: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Recreates the delta tables in the lakehouse based on the DAX expressions stored as model annotations in the Direct Lake semantic model.

    Parameters
    ----------
    dataset : str | UUID
        Name or ID of the semantic model.
    workspace : str | UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)
    icons.sll_tags.append("DirectLakeMigration")

    @retry(
        sleep_time=1,
        timeout_error_message=f"{icons.red_dot} Function timed out after 1 minute",
    )
    def dyn_connect():
        with connect_semantic_model(
            dataset=dataset_id, readonly=True, workspace=workspace_id
        ) as tom:

            tom.model

    dyn_connect()

    with connect_semantic_model(
        dataset=dataset_id, readonly=True, workspace=workspace_id
    ) as tom:
        for a in tom.model.Annotations:
            if any(a.Name == t.Name for t in tom.model.Tables):
                tName = a.Name
                query = a.Value

                if not query.startswith("EVALUATE"):
                    daxquery = "EVALUATE \n" + query
                else:
                    daxquery = query

                try:
                    df = fabric.evaluate_dax(
                        dataset=dataset_id,
                        dax_string=daxquery,
                        workspace=workspace_id,
                    )

                    # Update column names for non-field parameters
                    if query.find("NAMEOF") == -1:
                        for old_column_name in df.columns:
                            pattern = r"\[([^\]]+)\]"

                            matches = re.findall(pattern, old_column_name)
                            new_column_name = matches[0]
                            new_column_name = new_column_name.replace(" ", "")

                            df.rename(
                                columns={old_column_name: new_column_name},
                                inplace=True,
                            )

                            # Update data types for lakehouse columns
                            dataType = next(
                                str(c.DataType)
                                for c in tom.all_columns()
                                if c.Parent.Name == tName
                                and c.SourceColumn == new_column_name
                            )
                            # dfC_type = dfC[(dfC['Table Name'] == tName) & (dfC['Source'] == new_column_name)]
                            # dataType = dfC_type['Data Type'].iloc[0]

                            if dataType == "Int64":
                                df[new_column_name] = df[new_column_name].astype(int)
                            elif dataType in ["Decimal", "Double"]:
                                df[new_column_name] = df[new_column_name].astype(float)
                            elif dataType == "Boolean":
                                df[new_column_name] = df[new_column_name].astype(bool)
                            elif dataType == "DateTime":
                                df[new_column_name] = pd.to_datetime(
                                    df[new_column_name]
                                )
                            else:
                                df[new_column_name] = df[new_column_name].astype(str)
                    # else:
                    #    second_column_name = df.columns[1]
                    #    third_column_name = df.columns[2]
                    #    df[third_column_name] = df[third_column_name].astype(int)

                    # Remove calc columns from field parameters
                    #    mask = df[second_column_name].isin(dfC_filt['Full Column Name'])
                    #    df = df[~mask]

                    delta_table_name = tName.replace(" ", "_")
                    print(
                        f"{icons.in_progress} Refresh of the '{delta_table_name}' table within the lakehouse is in progress..."
                    )

                    save_as_delta_table(
                        dataframe=df,
                        table_name=delta_table_name,
                        write_mode="overwrite",
                    )

                    print(
                        f"{icons.green_dot} Calculated table '{tName}' has been refreshed as the '{delta_table_name.lower()}' table in the lakehouse."
                    )
                except Exception as e:
                    raise ValueError(
                        f"{icons.red_dot} Failed to create calculated table '{tName}' as a delta table in the lakehouse."
                    ) from e



import sempy.fabric as fabric
import pandas as pd
import json
import os
import copy
from anytree import Node, RenderTree
from powerbiclient import Report
from pyspark.sql.functions import col, flatten
from sempy_labs.report._generate_report import update_report_from_reportjson
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from sempy_labs._helper_functions import (
    resolve_report_id,
    language_validate,
    resolve_workspace_name_and_id,
    _decode_b64,
    resolve_dataset_id,
    _update_dataframe_datatypes,
    _base_api,
    _create_spark_session,
    _mount,
    resolve_workspace_id,
    resolve_item_name_and_id,
)
from typing import List, Optional, Union
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID


@log
def get_report_json(
    report: str,
    workspace: Optional[str | UUID] = None,
    save_to_file_name: Optional[str] = None,
) -> dict:
    """
    Gets the report.json file content of a Power BI report. This function only supports reports in the PBIR-Legacy format.

    This is a wrapper function for the following API: `Items - Get Report Definition <https://learn.microsoft.com/rest/api/fabric/report/items/get-report-definition>`_.

    Parameters
    ----------
    report : str | uuid.UUID
        Name or ID of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the report exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    save_to_file_name : str, default=None
        Specifying this parameter will save the report.json file to the lakehouse attached to the notebook with the file name of this parameter.

    Returns
    -------
    dict
        The report.json file for a given Power BI report.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (report_name, report_id) = resolve_item_name_and_id(
        item=report, type="Report", workspace=workspace_id
    )

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/reports/{report_id}/getDefinition",
        method="post",
        lro_return_json=True,
        status_codes=None,
    )
    report_json = None
    for part in result.get("definition", {}).get("parts", {}):
        if part.get("path") == "report.json":
            payload = part.get("payload")
            report_file = _decode_b64(payload)
            report_json = json.loads(report_file)

    if not report_json:
        raise ValueError(
            f"{icons.red_dot} Unable to retrieve report.json for the '{report_name}' report within the '{workspace_name}' workspace. This function only supports reports in the PBIR-Legacy format."
        )

    if save_to_file_name is not None:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} In order to save the report.json file, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
            )

        local_path = _mount()
        save_folder = f"{local_path}/Files"
        file_ext = ".json"
        if not save_to_file_name.endswith(file_ext):
            save_to_file_name = f"{save_to_file_name}{file_ext}"
        file_path = os.path.join(save_folder, save_to_file_name)
        with open(file_path, "w") as json_file:
            json.dump(report_json, json_file, indent=4)
        print(
            f"{icons.green_dot} The report.json file for the '{report}' report has been saved to the lakehouse attached to this notebook in this location: Files/'{save_to_file_name}'.\n\n"
        )

    return report_json


@log
def report_dependency_tree(workspace: Optional[str | UUID] = None):
    """
    Prints a dependency between reports and semantic models.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfR = fabric.list_reports(workspace=workspace_id)
    dfD = fabric.list_datasets(workspace=workspace_id)
    dfR = pd.merge(
        dfR,
        dfD[["Dataset ID", "Dataset Name"]],
        left_on="Dataset Id",
        right_on="Dataset ID",
        how="left",
    )
    dfR.rename(columns={"Name": "Report Name"}, inplace=True)
    dfR = dfR[["Report Name", "Dataset Name"]]

    report_icon = "\U0001f4f6"
    dataset_icon = "\U0001f9ca"
    workspace_icon = "\U0001f465"

    node_dict = {}
    rootNode = Node(workspace_name)
    node_dict[workspace_name] = rootNode
    rootNode.custom_property = f"{workspace_icon} "

    for _, r in dfR.iterrows():
        datasetName = r["Dataset Name"]
        reportName = r["Report Name"]
        parentNode = node_dict.get(datasetName)
        if parentNode is None:
            parentNode = Node(datasetName, parent=rootNode)
            node_dict[datasetName] = parentNode
        parentNode.custom_property = f"{dataset_icon} "

        child_node = Node(reportName, parent=parentNode)
        child_node.custom_property = f"{report_icon} "

    # Print the tree structure
    for pre, _, node in RenderTree(node_dict[workspace_name]):
        print(f"{pre}{node.custom_property}'{node.name}'")


@log
def clone_report(
    report: str,
    cloned_report: str,
    workspace: Optional[str | UUID] = None,
    target_workspace: Optional[str] = None,
    target_dataset: Optional[str] = None,
    target_dataset_workspace: Optional[str] = None,
):
    """
    Clones a Power BI report.

    This is a wrapper function for the following API: `Reports - Clone Report In Group <https://learn.microsoft.com/rest/api/power-bi/reports/clone-report-in-group>`_.

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    cloned_report : str
        Name of the new Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    target_workspace : str, default=None
        The name of the Fabric workspace to place the cloned report.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    target_dataset : str, default=None
        The name of the semantic model to be used by the cloned report.
        Defaults to None which resolves to the semantic model used by the initial report.
    target_dataset_workspace : str, default=None
        The workspace name in which the semantic model to be used by the report resides.
        Defaults to None which resolves to the semantic model used by the initial report.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfI = fabric.list_items(workspace=workspace_id, type="Report")
    dfI_filt = dfI[(dfI["Display Name"] == report)]

    if len(dfI_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{report}' report does not exist within the '{workspace_name}' workspace."
        )

    reportId = resolve_report_id(report, workspace_id)

    if target_workspace is None:
        target_workspace = workspace_name
        target_workspace_id = workspace_id
    else:
        target_workspace_id = resolve_workspace_id(workspace=target_workspace)

    if target_dataset is not None:
        if target_dataset_workspace is None:
            target_dataset_workspace = workspace_name
        target_dataset_id = resolve_dataset_id(target_dataset, target_dataset_workspace)

    if report == cloned_report and workspace_name == target_workspace:
        raise ValueError(
            f"{icons.warning} The 'report' and 'cloned_report' parameters have the same value of '{report}. The 'workspace' and 'target_workspace' have the same value of '{workspace_name}'. Either the 'cloned_report' or the 'target_workspace' must be different from the original report."
        )

    payload = {"name": cloned_report}
    if target_dataset is not None:
        payload["targetModelId"] = target_dataset_id
    if target_workspace != workspace_name:
        payload["targetWorkspaceId"] = target_workspace_id

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/reports/{reportId}/Clone",
        method="post",
        payload=payload,
    )
    print(
        f"{icons.green_dot} The '{report}' report has been successfully cloned as the '{cloned_report}' report within the '{target_workspace}' workspace."
    )


@log
def launch_report(report: str, workspace: Optional[str | UUID] = None):
    """
    Shows a Power BI report within a Fabric notebook.

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        An embedded Power BI report within the notebook.
    """

    from sempy_labs import resolve_report_id

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    report_id = resolve_report_id(report, workspace_id)
    report = Report(group_id=workspace_id, report_id=report_id)

    return report


@log
def list_report_pages(report: str, workspace: Optional[str | UUID] = None):
    """
    Shows the properties of all pages within a Power BI report.

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the pages within a Power BI report and their properties.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = pd.DataFrame(
        columns=["Page ID", "Page Name", "Hidden", "Width", "Height", "Visual Count"]
    )

    reportJson = get_report_json(report=report, workspace=workspace_id)

    for section in reportJson["sections"]:
        pageID = section.get("name")
        pageName = section.get("displayName")
        # pageFilters = section['filters']
        pageWidth = section.get("width")
        pageHeight = section.get("height")
        visualCount = len(section["visualContainers"])
        pageHidden = False
        pageConfig = section.get("config")
        pageConfigJson = json.loads(pageConfig)

        try:
            pageH = pageConfigJson["visibility"]
            if pageH == 1:
                pageHidden = True
        except Exception:
            pass

        new_data = {
            "Page ID": pageID,
            "Page Name": pageName,
            "Hidden": pageHidden,
            "Width": pageWidth,
            "Height": pageHeight,
            "Visual Count": visualCount,
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    column_map = {
        "Hidden": "bool",
        "Width": "int",
        "Height": "int",
        "Visual Count": "int",
    }

    _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    return df


@log
def list_report_visuals(report: str, workspace: Optional[str | UUID] = None):
    """
    Shows the properties of all visuals within a Power BI report.

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the visuals within a Power BI report and their properties.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    reportJson = get_report_json(report=report, workspace=workspace_id)

    df = pd.DataFrame(columns=["Page Name", "Page ID", "Visual ID", "Title"])

    for section in reportJson["sections"]:
        pageID = section["name"]
        pageName = section["displayName"]

        for visual in section["visualContainers"]:
            visualConfig = visual["config"]
            visualConfigJson = json.loads(visualConfig)
            visualID = visualConfigJson["name"]

            try:
                title = visualConfigJson["singleVisual"]["vcObjects"]["title"][0][
                    "properties"
                ]["text"]["expr"]["Literal"]["Value"]
                title = title[1:-1]
            except Exception:
                title = ""

            new_data = {
                "Page Name": pageName,
                "Page ID": pageID,
                "Visual ID": visualID,
                "Title": title,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


@log
def list_report_bookmarks(report: str, workspace: Optional[str | UUID] = None):
    """
    Shows the properties of all bookmarks within a Power BI report.

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the bookmarks within a Power BI report and their properties.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = pd.DataFrame(
        columns=[
            "Bookmark ID",
            "Bookmark Name",
            "Page ID",
            "Visual ID",
            "Visual Hidden",
        ]
    )

    reportJson = get_report_json(report=report, workspace=workspace_id)
    reportConfig = reportJson["config"]
    reportConfigJson = json.loads(reportConfig)

    try:
        for bookmark in reportConfigJson["bookmarks"]:
            bID = bookmark["name"]
            bName = bookmark["displayName"]
            rptPageId = bookmark["explorationState"]["activeSection"]

            for rptPg in bookmark["explorationState"]["sections"]:
                for vc in bookmark["explorationState"]["sections"][rptPg][
                    "visualContainers"
                ]:
                    vHidden = False
                    try:
                        hidden = bookmark["explorationState"]["sections"][rptPg][
                            "visualContainers"
                        ][vc]["singleVisual"]["display"]["mode"]
                        if hidden == "hidden":
                            vHidden = True
                    except Exception:
                        pass

            new_data = {
                "Bookmark ID": bID,
                "Bookmark Name": bName,
                "Page ID": rptPageId,
                "Visual ID": vc,
                "Visual Hidden": vHidden,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

        listPages = list_report_pages(report=report, workspace=workspace_id)

        df = pd.merge(df, listPages[["Page ID", "Page Name"]], on="Page ID", how="left")
        df = df[
            [
                "Bookmark ID",
                "Bookmark Name",
                "Page ID",
                "Page Name",
                "Visual ID",
                "Visual Hidden",
            ]
        ]

        return df

    except Exception:
        print(
            f"The '{report}' report within the '{workspace_name}' workspace has no bookmarks."
        )


@log
def translate_report_titles(
    report: str,
    languages: Union[str, List[str]],
    workspace: Optional[str | UUID] = None,
):
    """
    Dynamically generates new Power BI reports which have report titles translated into the specified language(s).

    Parameters
    ----------
    report : str
        Name of the Power BI report.
    languages : str, List[str]
        The language code(s) in which to translate the report titles.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """
    from synapse.ml.services import Translate

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    if isinstance(languages, str):
        languages = [languages]

    for lang in languages:
        language_validate(lang)

    reportJson = get_report_json(report=report, workspace=workspace_id)
    dfV = list_report_visuals(report=report, workspace=workspace_id)
    spark = _create_spark_session()
    df = spark.createDataFrame(dfV)
    columnToTranslate = "Title"

    translate = (
        Translate()
        .setTextCol(columnToTranslate)
        .setToLanguage(languages)
        .setOutputCol("translation")
        .setConcurrency(5)
    )

    transDF = (
        translate.transform(df)
        .withColumn("translation", flatten(col("translation.translations")))
        .withColumn("translation", col("translation.text"))
        .select("Visual ID", columnToTranslate, "translation")
    )

    df_panda = transDF.toPandas()

    i = 0
    for lang in languages:
        # Clone report
        language = language_validate(lang)
        clonedReportName = f"{report}_{language}"

        dfRep = fabric.list_reports(workspace=workspace_id)
        dfRep_filt = dfRep[
            (dfRep["Name"] == clonedReportName)
            & (dfRep["Report Type"] == "PowerBIReport")
        ]

        if len(dfRep_filt) > 0:
            print(
                f"{icons.yellow_dot} The '{clonedReportName}' report already exists in the '{workspace_name} workspace."
            )
        else:
            clone_report(
                report=report, cloned_report=clonedReportName, workspace=workspace_id
            )
            print(
                f"{icons.green_dot} The '{clonedReportName}' report has been created via clone in the '{workspace_name} workspace."
            )

        rptJsonTr = copy.deepcopy(reportJson)

        # Update report json file
        for section in rptJsonTr["sections"]:
            for visual in section["visualContainers"]:
                visualConfig = visual["config"]
                visualConfigJson = json.loads(visualConfig)
                visualID = visualConfigJson["name"]

                df_filt = df_panda[
                    (df_panda["Visual ID"] == visualID) & (df_panda["Title"] != "")
                ]

                if len(df_filt) == 1:
                    tr = df_filt["translation"].str[i].iloc[0]
                    if len(tr) > 0:
                        prop = visualConfigJson["singleVisual"]["vcObjects"]["title"][
                            0
                        ]["properties"]["text"]["expr"]["Literal"]
                        prop["Value"] = f"'{tr}'"

                visual["config"] = json.dumps(visualConfigJson)

        i += 1

        # Post updated report json file to cloned report
        update_report_from_reportjson(
            report=clonedReportName, report_json=rptJsonTr, workspace=workspace_id
        )
        print(
            f"{icons.green_dot} The visual titles within the '{clonedReportName}' report within the '{workspace_name}' have been translated into '{language}' accordingly."
        )



from ._items import (
    list_custom_pools,
    delete_custom_pool,
    update_custom_pool,
    update_spark_settings,
    create_custom_pool,
    get_spark_settings,
)

__all__ = [
    "list_custom_pools",
    "delete_custom_pool",
    "update_custom_pool",
    "update_spark_settings",
    "create_custom_pool",
    "get_spark_settings",
]



import pandas as pd
from uuid import UUID
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    _is_valid_uuid,
)
from sempy._utils._log import log


@log
def list_sensitivity_labels(user: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Get a list of `sensitivity label <https://learn.microsoft.com/graph/api/resources/security-sensitivitylabel>`_ objects associated with a user or organization.

    This is a wrapper function for the following API: `List sensitivityLabels https://learn.microsoft.com/graph/api/security-informationprotection-list-sensitivitylabels>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : str | uuid.UUID
        The user ID or user principal name.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of `sensitivity label <https://learn.microsoft.com/graph/api/resources/security-sensitivitylabel>`_ objects associated with a user or organization.
    """
    from sempy_labs.graph import resolve_user_id

    url = "/security/informationProtection/sensitivityLabels"

    if user is not None:
        user_id = resolve_user_id(user=user)
        url = f"users/{user_id}{url}"

    result = _base_api(request=url, client="graph").json()

    columns = {
        "Sensitivity Label Id": "str",
        "Sensitivity Label Name": "str",
        "Description": "str",
        "Color": "str",
        "Sensitivity": "int",
        "Tooltip": "str",
        "Is Active": "bool",
        "Is Appliable": "bool",
        "Has Protection": "bool",
        "Parent Sensitivity Label Id": "str",
        "Parent Sensitivity Label Name": "str",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    for item in result.get("value", []):
        row = {
            "Sensitivity Label Id": item.get("id"),
            "Sensitivity Label Name": item.get("name"),
            "Description": item.get("description"),
            "Color": item.get("color"),
            "Sensitivity": item.get("sensitivity"),
            "Tooltip": item.get("tooltip"),
            "Is Active": item.get("isActive"),
            "Is Appliable": item.get("isAppliable"),
            "Has Protection": item.get("hasProtection"),
            "Parent Sensitivity Label Id": (
                item.get("parent", {}).get("id") if item.get("parent") else None
            ),
            "Parent Sensitivity Label Name": (
                item.get("parent", {}).get("name") if item.get("parent") else None
            ),
        }
        rows.append(row)

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def resolve_sensitivity_label_id(
    label: str | UUID, user: Optional[str | UUID] = None
) -> UUID | None:
    """
    Resolve a sensitivity label name or ID to its corresponding sensitivity label ID.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    label : str | uuid.UUID
        The name or ID of the sensitivity label.
    user : str | uuid.UUID, default=None
        The user ID or user principal name.

    Returns
    -------
    uuid.UUID | None
        The ID of the sensitivity label if found, otherwise None.
    """

    if _is_valid_uuid(label):
        return str(label)

    df = list_sensitivity_labels(user=user)

    if df.empty:
        return None

    # Try to find the label by name
    label_row = df[df["Sensitivity Label Name"] == label]
    if not label_row.empty:
        return label_row["Sensitivity Label Id"].iloc[0]

    return None



from typing import Optional
from uuid import UUID
from sempy._utils._log import log


@log
def get_shared_expression(
    lakehouse: Optional[str] = None, workspace: Optional[str | UUID] = None
) -> str:
    """
    Dynamically generates the M expression used by a Direct Lake model for a given lakehouse.

    Parameters
    ----------
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        Shows the expression which can be used to connect a Direct Lake semantic model to its SQL Endpoint.
    """
    from sempy_labs.directlake._generate_shared_expression import (
        generate_shared_expression,
    )

    return generate_shared_expression(
        item_name=lakehouse, item_type="Lakehouse", workspace=workspace
    )



from ._items import (
    list_managed_private_endpoint_fqdns,
    list_managed_private_endpoints,
    delete_managed_private_endpoint,
    create_managed_private_endpoint,
)

__all__ = [
    "list_managed_private_endpoint_fqdns",
    "list_managed_private_endpoints",
    "delete_managed_private_endpoint",
    "create_managed_private_endpoint",
]



from ._functions import (
    get_package_tmdl,
    get_package_functions,
    add_package_to_semantic_model,
    remove_package_from_semantic_model,
    update_package_in_semantic_model,
)

__all__ = [
    "get_package_tmdl",
    "get_package_functions",
    "add_package_to_semantic_model",
    "remove_package_from_semantic_model",
    "update_package_in_semantic_model",
]



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    delete_item,
    _create_dataframe,
    create_item,
    resolve_workspace_id,
    resolve_item_id,
    _decode_b64,
)
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log
import json


@log
def list_eventstreams(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the eventstreams within a workspace.

    This is a wrapper function for the following API: `Items - List Eventstreams <https://learn.microsoft.com/rest/api/fabric/environment/items/list-eventstreams>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the eventstreams within a workspace.
    """

    columns = {
        "Eventstream Name": "string",
        "Eventstream Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams", uses_pagination=True
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Eventstream Name": v.get("displayName"),
                    "Eventstream Id": v.get("id"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def create_eventstream(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric eventstream.

    This is a wrapper function for the following API: `Items - Create Eventstream <https://learn.microsoft.com/rest/api/fabric/environment/items/create-eventstream>`_.

    Parameters
    ----------
    name: str
        Name of the eventstream.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="Eventstream", workspace=workspace
    )


@log
def delete_eventstream(
    eventstream: str | UUID, workspace: Optional[str | UUID] = None, **kwargs
):
    """
    Deletes a Fabric eventstream.

    This is a wrapper function for the following API: `Items - Delete Eventstream <https://learn.microsoft.com/rest/api/fabric/environment/items/delete-eventstream>`_.

    Parameters
    ----------
    eventstream: str | uuid.UUID
        Name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if "name" in kwargs:
        eventstream = kwargs["name"]
        print(
            f"{icons.warning} The 'name' parameter is deprecated. Please use 'eventstream' instead."
        )

    delete_item(item=eventstream, type="Eventstream", workspace=workspace)


@log
def get_eventstream_definition(
    eventstream: str | UUID,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
    return_dataframe: bool = False,
) -> dict:

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=eventstream, type="Eventstream", workspace=workspace)

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams/{item_id}/getDefinition",
        method="post",
        client="fabric_sp",
        status_codes=None,
        lro_return_json=True,
    )

    if decode:
        definition = {"definition": {"parts": []}}

        for part in result.get("definition", {}).get("parts", []):
            path = part.get("path")
            payload = json.loads(_decode_b64(part.get("payload")))
            definition["definition"]["parts"].append({"path": path, "payload": payload})
    else:
        definition = result.copy()

    if return_dataframe:
        df = pd.DataFrame(definition["definition"]["parts"])
        df.columns = ["Path", "Payload", "Payload Type"]
        return df
    else:
        return definition


@log
def list_eventstream_destinations(
    eventstream: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Lists the destinations of the specified eventstream.

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the destinations of the eventstream.
    """

    definition = get_eventstream_definition(
        eventstream=eventstream, workspace=workspace
    )

    columns = {
        "Destination Id": "string",
        "Destination Name": "string",
        "Destination Type": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for part in definition.get("definition").get("parts"):
        payload = part.get("payload")
        if part.get("path") == "eventstream.json":
            destinations = payload.get("destinations")
            for d in destinations:
                rows.append(
                    {
                        "Destination Id": d.get("id"),
                        "Destination Name": d.get("name"),
                        "Destination Type": d.get("type"),
                    }
                )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_eventstream_sources(
    eventstream: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Lists the destinations of the specified eventstream.

    Parameters
    ----------
    eventstream : str | uuid.UUID
        The name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the destinations of the eventstream.
    """

    definition = get_eventstream_definition(
        eventstream=eventstream, workspace=workspace
    )

    columns = {
        "Source Id": "string",
        "Source Name": "string",
        "Source Type": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for part in definition.get("definition").get("parts"):
        payload = part.get("payload")
        if part.get("path") == "eventstream.json":
            sources = payload.get("sources")
            for s in sources:
                rows.append(
                    {
                        "Source Id": s.get("id"),
                        "Source Name": s.get("name"),
                        "Source Type": s.get("type"),
                    }
                )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



from typing import Optional
from uuid import UUID
import pandas as pd
from sempy_labs._helper_functions import (
    _create_dataframe,
    _base_api,
    resolve_workspace_id,
    resolve_item_id,
)
from sempy._utils._log import log


@log
def list_data_access_roles(
    item: str | UUID, type: str, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns a list of OneLake roles.

    This is a wrapper function for the following API: `OneLake Data Access Security - List Data Access Roles <https://learn.microsoft.com/rest/api/fabric/core/onelake-data-access-security/list-data-access-roles>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item.
    type : str
        The type of the item.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of OneLake roles.
    """

    columns = {
        "Role Name": "string",
        "Effect": "string",
        "Attribute Name": "string",
        "Attribute Values": "string",
        "Item Access": "string",
        "Source Path": "string",
    }

    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(item=item, type=type, workspace=workspace_id)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/dataAccessRoles",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for role in r.get("value", []):
            name = role.get("name")

            # Loop through members first (since they are crucial)
            members = role.get("members", {}).get("fabricItemMembers", [])
            if not members:
                members = [{}]  # if no members exist, still create at least one row

            for member in members:
                item_access = member.get("itemAccess", [])
                source_path = member.get("sourcePath")

                # Loop through decision rules
                for rule in role.get("decisionRules", []):
                    effect = rule.get("effect")

                    # Loop through permissions
                    for perm in rule.get("permission", []):
                        attr_name = perm.get("attributeName")
                        attr_values = perm.get("attributeValueIncludedIn", [])

                        rows.append(
                            {
                                "Role Name": name,
                                "Effect": effect,
                                "Attribute Name": attr_name,
                                "Attribute Values": ", ".join(attr_values),
                                "Item Access": ", ".join(item_access),
                                "Source Path": source_path,
                            }
                        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



import re
import json
import difflib
from collections import defaultdict


def color_text(text, color_code):
    return f"\033[{color_code}m{text}\033[0m"


def stringify(payload):
    try:
        if isinstance(payload, list):
            return (
                "[\n" + ",\n".join(f"  {json.dumps(item)}" for item in payload) + "\n]"
            )
        return json.dumps(payload, indent=2, sort_keys=True)
    except Exception:
        return str(payload)


def extract_top_level_group(path):
    # For something like: resourcePackages[1].items[1].name → resourcePackages[1].items[1]
    segments = re.split(r"\.(?![^[]*\])", path)  # split on dots not in brackets
    return ".".join(segments[:-1]) if len(segments) > 1 else segments[0]


def get_by_path(obj, path):
    """Navigate into nested dict/list based on a dot/bracket path like: a.b[1].c"""
    tokens = re.findall(r"\w+|\[\d+\]", path)
    for token in tokens:
        if token.startswith("["):
            index = int(token[1:-1])
            obj = obj[index]
        else:
            obj = obj.get(token)
    return obj


def deep_diff(d1, d2, path=""):
    diffs = []
    if isinstance(d1, dict) and isinstance(d2, dict):
        keys = set(d1) | set(d2)
        for key in sorted(keys):
            new_path = f"{path}.{key}" if path else key
            if key not in d1:
                diffs.append(("+", new_path, None, d2[key]))
            elif key not in d2:
                diffs.append(("-", new_path, d1[key], None))
            else:
                diffs.extend(deep_diff(d1[key], d2[key], new_path))
    elif isinstance(d1, list) and isinstance(d2, list):
        min_len = min(len(d1), len(d2))
        list_changed = False
        for i in range(min_len):
            if d1[i] != d2[i]:
                list_changed = True
                break
        if list_changed or len(d1) != len(d2):
            diffs.append(("~", path, d1, d2))
    elif d1 != d2:
        diffs.append(("~", path, d1, d2))
    return diffs


def diff_parts(d1, d2):

    def build_path_map(parts):
        return {part["path"]: part["payload"] for part in parts}

    try:
        paths1 = build_path_map(d1)
    except Exception:
        paths1 = d1
    try:
        paths2 = build_path_map(d2)
    except Exception:
        paths2 = d2
    all_paths = set(paths1) | set(paths2)

    for part_path in sorted(all_paths):
        p1 = paths1.get(part_path)
        p2 = paths2.get(part_path)

        if p1 is None:
            print(color_text(f"+ {part_path}", "32"))  # Green
            continue
        elif p2 is None:
            print(color_text(f"- {part_path}", "31"))  # Red
            continue
        elif p1 == p2:
            continue

        if p1 is None or p2 is None:
            print(
                color_text(f"+ {part_path}", "32")
                if p2 and not p1
                else color_text(f"- {part_path}", "31")
            )
            continue

        # Header for the changed part
        print(color_text(f"~ {part_path}", "33"))

        # Collect diffs
        diffs = deep_diff(p1, p2)
        # If the diff is only a change of a whole list (like appending to a list), group it under its key
        merged_list_diffs = []
        for change_type, full_path, old_val, new_val in diffs:
            if (
                change_type == "~"
                and isinstance(old_val, list)
                and isinstance(new_val, list)
            ):
                merged_list_diffs.append((change_type, full_path, old_val, new_val))

        # Replace individual item diffs with unified list diff
        if merged_list_diffs:
            diffs = merged_list_diffs

        # Group diffs by common parent path (e.g. items[1])
        grouped = defaultdict(list)
        for change_type, full_path, old_val, new_val in diffs:
            group_path = extract_top_level_group(full_path)
            grouped[group_path].append((change_type, full_path, old_val, new_val))

        # Print each group once with unified diff for the full substructure
        for group_path in sorted(grouped):
            print("  " + color_text(f"~ {group_path}", "33"))

            try:
                old_group = get_by_path(p1, group_path)
                new_group = get_by_path(p2, group_path)
            except Exception:
                old_group = new_group = None

            # Skip showing diffs for empty/null groups
            if isinstance(old_group, dict) and isinstance(new_group, dict):
                old_keys = set(old_group.keys())
                new_keys = set(new_group.keys())

                for key in sorted(old_keys - new_keys):
                    print(
                        "  "
                        + color_text(f"- {key}: {json.dumps(old_group[key])}", "31")
                    )
                for key in sorted(new_keys - old_keys):
                    print(
                        "  "
                        + color_text(f"+ {key}: {json.dumps(new_group[key])}", "32")
                    )
                for key in sorted(old_keys & new_keys):
                    if old_group[key] != new_group[key]:
                        print("  " + color_text(f"~ {key}:", "33"))
                        old_val_str = stringify(old_group[key]).splitlines()
                        new_val_str = stringify(new_group[key]).splitlines()
                        for line in difflib.unified_diff(
                            old_val_str,
                            new_val_str,
                            fromfile="old",
                            tofile="new",
                            lineterm="",
                        ):
                            if line.startswith("@@"):
                                print("    " + color_text(line, "36"))
                            elif line.startswith("-") and not line.startswith("---"):
                                print("    " + color_text(line, "31"))
                            elif line.startswith("+") and not line.startswith("+++"):
                                print("    " + color_text(line, "32"))
            elif old_group is None and new_group is not None:
                if isinstance(new_group, dict):
                    # print all added keys
                    for key, val in new_group.items():
                        print("  " + color_text(f"+ {key}: {json.dumps(val)}", "32"))
                elif isinstance(new_group, list):
                    old_str = []
                    new_str = stringify(new_group).splitlines()
                    for line in difflib.unified_diff(
                        old_str, new_str, fromfile="old", tofile="new", lineterm=""
                    ):
                        if line.startswith("@@"):
                            print("  " + color_text(line, "36"))
                        elif line.startswith("-") and not line.startswith("---"):
                            print("  " + color_text(line, "31"))
                        elif line.startswith("+") and not line.startswith("+++"):
                            print("  " + color_text(line, "32"))
                else:
                    print("  " + color_text(f"+ {json.dumps(new_group)}", "32"))

            elif new_group is None and old_group is not None:
                if isinstance(old_group, dict):
                    # print all removed keys
                    for key, val in old_group.items():
                        print("  " + color_text(f"- {key}: {json.dumps(val)}", "31"))
                elif isinstance(old_group, list):
                    old_str = stringify(old_group).splitlines()
                    new_str = []
                    for line in difflib.unified_diff(
                        old_str, new_str, fromfile="old", tofile="new", lineterm=""
                    ):
                        if line.startswith("@@"):
                            print("  " + color_text(line, "36"))
                        elif line.startswith("-") and not line.startswith("---"):
                            print("  " + color_text(line, "31"))
                        elif line.startswith("+") and not line.startswith("+++"):
                            print("  " + color_text(line, "32"))
                else:
                    print("  " + color_text(f"- {json.dumps(old_group)}", "31"))
            else:
                old_str = stringify(old_group).splitlines()
                new_str = stringify(new_group).splitlines()

                for line in difflib.unified_diff(
                    old_str, new_str, fromfile="old", tofile="new", lineterm=""
                ):
                    if line.startswith("@@"):
                        print("  " + color_text(line, "36"))
                    elif line.startswith("-") and not line.startswith("---"):
                        print("  " + color_text(line, "31"))
                    elif line.startswith("+") and not line.startswith("+++"):
                        print("  " + color_text(line, "32"))



import sempy_labs._icons as icons
from typing import Optional, Literal
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_lakehouse_name_and_id,
    _base_api,
    _mount,
    resolve_workspace_name,
    _is_valid_uuid,
)
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from uuid import UUID
from sempy._utils._log import log
import sempy.fabric as fabric


@log
def download_report(
    report: str | UUID,
    file_name: Optional[str] = None,
    download_type: Literal["LiveConnect", "IncludeModel"] = "LiveConnect",
    workspace: Optional[str | UUID] = None,
):
    """
    Downloads the specified report from the specified workspace to a Power BI .pbix file.

    This is a wrapper function for the following API: `Reports - Export Report In Group <https://learn.microsoft.com/rest/api/power-bi/reports/export-report-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    report: str | uuid.UUID
        Name or ID of the Power BI report or Paginated Report.
    file_name : str, default=None
        Name of the .pbix or .rdl file to be saved. Do not include the file extension.
        Defaults to None which resolves to the name of the report.
    download_type : typing.Literal["LiveConnect", "IncludeModel"], default="LiveConnect"
        The type of download. Valid values are "LiveConnect" and "IncludeModel". This parameter is ignored for Paginated Reports.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if not lakehouse_attached():
        raise ValueError(
            f"{icons.red_dot} A lakehouse must be attached to the notebook."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id()
    lakehouse_workspace = resolve_workspace_name()

    dfI = fabric.list_items(workspace=workspace_id)
    dfI_filt = dfI[dfI["Type"].isin(["Report", "PaginatedReport"])]
    if _is_valid_uuid(report):
        df = dfI_filt[dfI_filt["Id"] == str(report)]
    else:
        df = dfI_filt[dfI_filt["Display Name"] == report]

    if df.empty:
        raise ValueError(
            f"{icons.red_dot} The report '{report}' was not found in the '{workspace_name}' workspace."
        )

    report_name = df["Display Name"].iloc[0]
    report_id = df["Id"].iloc[0]
    report_type = df["Type"].iloc[0]

    download_types = ["LiveConnect", "IncludeModel"]
    if download_type not in download_types and report_type != "PaginatedReport":
        raise ValueError(
            f"{icons.red_dot} Invalid download_type parameter. Valid options: {download_types}."
        )

    file_name = file_name or report_name

    base_url = f"v1.0/myorg/groups/{workspace_id}/reports/{report_id}/Export"
    extension = ".pbix"
    if report_type == "PaginatedReport":
        response = _base_api(
            request=base_url,
            client="fabric_sp",
        )
        extension = ".rdl"
    else:
        response = _base_api(
            request=f"{base_url}?downloadType={download_type}",
            client="fabric_sp",
        )

    # Save file to the attached lakehouse
    local_path = _mount()
    save_file = f"{local_path}/Files/{file_name}{extension}"
    with open(save_file, "wb") as file:
        file.write(response.content)

    print(
        f"{icons.green_dot} The '{report_name}' {report_type} within the '{workspace_name}' workspace has been exported as the '{file_name}' file in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
    )
